#include "hcsr04.h"
#include "PinChangeInt.h"
#include <PinChangeIntConfig.h>
#include "Arduino.h"
sonar * sonars;

void sonars_init(sonar* sonar_array,int * sonar_echo_pins){
	sonars=sonar_array;
  for (int i=0;i<N_SONARS;i++){
    sonar *temp = &sonars[i];
    temp->trig_pin=common_trigger_pin;
    temp->echo_pin=sonar_echo_pins[i];
    temp->echo_start_time=0;
    temp->triggered=false;
    temp->echo_started=false;
    temp->distance_cm=0;
    temp->valid=false;
    
    pinMode(temp->trig_pin,OUTPUT);
    pinMode(temp->echo_pin,INPUT);
  }
  
  PCintPort::attachInterrupt(sonar_echo_pins[0], echo_isr1, CHANGE);
  PCintPort::attachInterrupt(sonar_echo_pins[1], echo_isr2, CHANGE);
  PCintPort::attachInterrupt(sonar_echo_pins[2], echo_isr3, CHANGE);
  PCintPort::attachInterrupt(sonar_echo_pins[3], echo_isr4, CHANGE);
  PCintPort::attachInterrupt(sonar_echo_pins[4], echo_isr5, CHANGE);
  PCintPort::attachInterrupt(sonar_echo_pins[5], echo_isr6, CHANGE);
  
  pinMode(common_trigger_pin,OUTPUT);
}

void sonars_trigger(){
  digitalWrite(common_trigger_pin,HIGH);
  delayMicroseconds(20);
  digitalWrite(common_trigger_pin,LOW);
  
  sonar *temp;
  for (int i=0;i<N_SONARS;i++){
    temp= &sonars[i];
    temp->valid=false;
    temp->triggered=true;
	temp->echo_started=false;
  }
}

void interrupt_handler(int n){
  //Serial.print("INTERRUPT ");
  //Serial.println(n);
  if (n>N_SONARS || n<1){
	  return;
  }
  sonar *temp = &sonars[n-1];
  
  if (temp->triggered==false){
    temp->echo_started=false; // Kan fjernes?
    temp->valid=false;
    return;
  }
  
  if (temp->echo_started==false && digitalRead(temp->echo_pin)==HIGH){
    temp->echo_started=true;
    temp->echo_start_time=micros();
  }
  else if (temp->echo_started==true && digitalRead(temp->echo_pin)==LOW){
    //double current_time=(double)micros();
	//double start_time = (double)temp->echo_start_time;
	//roundtrip=current_time-start_time;
	double roundtrip=(double)(micros()-temp->echo_start_time-5); //5 us overhaed
    temp->distance_cm=roundtrip/44.95;
    temp->echo_started=false;
    temp->valid=true;
    temp->triggered=false;
  }
  else{
    //something wrong has happened
    //Serial.println("ERROR in interrupt_handler");
    return;
  }
}
    
void echo_isr1(){
  interrupt_handler(1);
}

void echo_isr2(){
  interrupt_handler(2);
}

void echo_isr3(){
  interrupt_handler(3);
}

void echo_isr4(){
  interrupt_handler(4);
}

void echo_isr5(){
  interrupt_handler(5);
}

void echo_isr6(){
  interrupt_handler(6);
}

void sonar_get_values(double * returned){
	sonar *temp;
	for (int i=0;i<N_SONARS;i++){
		temp = &sonars[i];
		if (temp->valid){
		  returned[i]=temp->distance_cm;
		}
		else {
		  returned[i]=-1;
		}
	}
}

    
  
  
  
