#include <Servo.h>
#include <Wire.h>
#include <I2Cdev.h>
#include <helper_3dmath.h>
#include <MPU6050_6Axis_MotionApps20.h>
#include <PID_v1.h>
#include <hcsr04.h>
#include "uart_datagram.h"

//#define DEBUG_PRINTING
//#define DEBUG_UART
//#define PRODUCTION_UART_PRINT

/* Serial stuff
*
*/
String inputString = "";         // a string to hold incoming data
boolean stringComplete = false;  // whether the string is complete
unsigned long last_received_time_ms=0;
#define UART_TIMEOUT_LIMIT_MS 500 //500 ms

/* Program flow
* 
*/
unsigned long cycle_interval_ms=100; // 10 hz = 100ms
unsigned long next_execution_time=0;


/* Sonar stuff 
*
*/
int sonar_echo_pins[N_SONARS] = {A0,A1,A2,A3,7,8};
// A6 and A7 does not work with interrupt because they are somehow not on a port on the AVR
sonar sonar_list[N_SONARS];
double sonar_values[N_SONARS];


/* ESC configuration
 *
 */
#define ESC_MIN 1000 //us for servo
#define ESC_MAX 2000
#define ESC_TAKEOFF_OFFSET 0
#define ESC_ARM_DELAY 50 //set this up to 5000 later

//pin outputs to Flight Controller
#define FC_PITCH 9 
#define FC_ROLL 6 
#define FC_YAW 5 
#define FC_THROTT 3 //RESET WOOT. Kanskje dette bare er D9,D6,D5,D3 ? ja tror det


/* Flight controller stuff
* 
*/
int pitch_sp, roll_sp, yaw_sp, thrott_sp;                    //velocities
//Servo objects for 50hz 1000-2000 us PWM
Servo pitch,roll,yaw,throttle;


/*  MPU6050 IMU variables
 *
 */
MPU6050 mpu;                           // mpu interface object
uint8_t mpuIntStatus;                  // mpu statusbyte
uint8_t devStatus;                     // device status    
uint16_t packetSize;                   // estimated packet size  
uint16_t fifoCount;                    // fifo buffer size   
uint8_t fifoBuffer[64];         // fifo buffer 
boolean mpu_calibration_complete = false;
#define mpu_interrupt_nr 0 //pin D2
Quaternion q;                          // quaternion for mpu output
VectorFloat gravity;                   // gravity vector for ypr
float ypr[3] = {0.0f,0.0f,0.0f};       // yaw pitch roll values
float yprLast[3] = {0.0f, 0.0f, 0.0f};
int16_t gx,gy,gz;
volatile bool mpuInterrupt = false;    //interrupt flag
boolean interruptLock = false;



/* Setup code
* 
*/
int counter=0;
void setup(){
  
  Serial.begin(115200);                 
  inputString.reserve(200);
  Serial.flush();
  //Serial.println("Hello world");
  sonars_init(sonar_list,sonar_echo_pins);
  initMPU();
  initESCs();
  next_execution_time=millis()+cycle_interval_ms;
}


/* Main loop
*
*/
void loop(){
  counter++;
  while(millis()<next_execution_time){
    //wait for next cycle
  }
  // Update timer variables
  next_execution_time=millis()+cycle_interval_ms;

  //IMU stuff
  if(mpuInterrupt){
    if (mpu_calibration_complete==false){
      mpu_calibrate();
      mpu_calibration_complete=true;
    }
   else { 
      getYPR();
   }
  }
  
  // Read sonar values
  sonar_get_values(sonar_values);
  
  // Serial send
  send_serial_data();
  // Serial read
  read_serial();
  // Update Flight controller
  updateFC();
  
  // Trigger sonars for next round
  sonars_trigger();
  
  
  
  // Debug print
#ifdef DEBUG_PRINTING
/*
      // ypr
      Serial.print(rad_to_deg(ypr[0]));Serial.print('\t');
      Serial.print(rad_to_deg(ypr[1]));Serial.print('\t');
      Serial.print(rad_to_deg(ypr[2]));Serial.print('\t');
      // gyro
      Serial.print("Gyro readings:\t");
      Serial.print(gx);Serial.print('\t');
      Serial.print(gy);Serial.print('\t');
      Serial.print(gz);Serial.print('\t');
      */
      //sonar
      Serial.print("Sonar values:");Serial.print('\t');
      for (int i=0;i<N_SONARS;i++){
        Serial.print(sonar_values[i]);Serial.print('\t');
        if(sonar_values[i]==-1){
          Serial.print(OMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFGOMFG};
        }  
      }
      
      /*
      //FC
      Serial.print("FC values:");Serial.print('\t');
      Serial.print(pitch_sp);Serial.print('\t');
      Serial.print(roll_sp);Serial.print('\t');
      Serial.print(yaw_sp);Serial.print('\t');
      Serial.print(thrott_sp);Serial.print('\t');
      */
      //newline
      Serial.println();
#endif
 }


void send_serial_data(){
  #ifdef PRODUCTION_UART_PRINT
  
  // Roll / Pitch / Yaw estimates
  Serial.print(PITCH);Serial.print(':');Serial.print(rad_to_deg(ypr[1]));Serial.print(';');  
  Serial.print(ROLL);Serial.print(':');Serial.print(rad_to_deg(ypr[2]));Serial.print(';');
  Serial.print(YAW);Serial.print(':');Serial.print(rad_to_deg(ypr[0]));Serial.print(';');
  // Gyro readings
  Serial.print(GYRO_X);Serial.print(':');Serial.print(gx);Serial.print(';');
  Serial.print(GYRO_Y);Serial.print(':');Serial.print(gy);Serial.print(';');
  Serial.print(GYRO_Z);Serial.print(':');Serial.print(gz);Serial.print(';');
  
  //Sonar values
  Serial.print(SONAR_FRONT);Serial.print(':');Serial.print(sonar_values[0]);Serial.print(';');
  Serial.print(SONAR_BACK);Serial.print(':');Serial.print(sonar_values[1]);Serial.print(';');
  Serial.print(SONAR_RIGHT);Serial.print(':');Serial.print(sonar_values[2]);Serial.print(';');
  Serial.print(SONAR_LEFT);Serial.print(':');Serial.print(sonar_values[3]);Serial.print(';');
  Serial.print(SONAR_DOWN);Serial.print(':');Serial.print(sonar_values[4]);Serial.print(';');
  Serial.print(SONAR_UP);Serial.print(':');Serial.print(sonar_values[5]);Serial.print(';');
  
  // Send newline
  Serial.print('\n');
  #endif
  
  //TODO: send status stuff
}
  
  
void on_uart_timeout(){
  //Serial.print("ERROR:UART TIMEOUT");
}

void read_serial(){
  if (Serial.available()<1){
    //no data ready, check for timeout
    if (millis()>last_received_time_ms+UART_TIMEOUT_LIMIT_MS){
      on_uart_timeout();
    }
    return;
  }
    while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read(); 
    // add it to the inputString:
    inputString += inChar;
    // Look for end-char? 
  }
  process_input_string(inputString);
  //clear buffer
  inputString = "";
  last_received_time_ms=millis();
}

void process_input_string(String s){
  int cur_index=0;
  int i=0;
  String type;
  String value;
  while(1){
    //extract type:
    i=s.indexOf(':',cur_index);
    if (i==-1)break;
    type=s.substring(cur_index,i);
    //extract value
    cur_index=i+1;
    i=s.indexOf(';',cur_index);
    if (i==-1)break;
    value=s.substring(cur_index,i);
    cur_index=i+1;
    
    // Take action based on received data
    if (type.equalsIgnoreCase(PITCH))pitch_sp=value.toInt();
    if (type.equalsIgnoreCase(ROLL))roll_sp=value.toInt();
    if (type.equalsIgnoreCase(YAW))yaw_sp=value.toInt();
    if (type.equalsIgnoreCase(THROTTLE))thrott_sp=value.toInt();
    
    #ifdef DEBUG_PRINTING 
    Serial.println(s);
    Serial.println(type);
    Serial.println(value);
    #endif    
  }
}

void fc_get_sp(){
  if (counter<ESC_MIN)counter=ESC_MIN;
  if (counter>ESC_MAX)counter=ESC_MIN;
  pitch_sp=counter;
  roll_sp=counter;
  yaw_sp=counter;
  thrott_sp=counter;
}

float rad_to_deg(float rad){
  return 180*rad/3.14;
}

float deg_ro_rad(float deg){
  return 3.14*deg/180;
}

// Get yaw pitch roll from IMU
void getYPR(){
    mpuInterrupt = false;
    mpuIntStatus = mpu.getIntStatus();
    fifoCount = mpu.getFIFOCount();
    
    if((mpuIntStatus & 0x10) || fifoCount >= 1024){ 
      mpu.resetFIFO(); 
    }else if(mpuIntStatus & 0x02){
      while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();
      mpu.getFIFOBytes(fifoBuffer, packetSize);
      fifoCount -= packetSize;
      mpu.dmpGetQuaternion(&q, fifoBuffer);
      mpu.dmpGetGravity(&gravity, &q);
      mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);      
      mpu.getRotation(&gx,&gy,&gz);
    }
}

//Send pwm to flight controller
void updateFC(){
  
  pitch.writeMicroseconds(pitch_sp);
  roll.writeMicroseconds(yaw_sp);
  yaw.writeMicroseconds(roll_sp);
  throttle.writeMicroseconds(thrott_sp);
}

//called by IMU interrupt routine
void dmpDataReady() {  
    mpuInterrupt = true;
    //Serial.println("dmpDataReady");
}

void initMPU(){  
  Wire.begin();
  mpu.initialize();
  devStatus = mpu.dmpInitialize();
  if(devStatus == 0){
  
    mpu.setDMPEnabled(true);
    attachInterrupt(0, dmpDataReady, RISING);
    mpuIntStatus = mpu.getIntStatus();
    packetSize = mpu.dmpGetFIFOPacketSize();
    
  }
}

void mpu_calibrate(){
  
  int16_t gx,gy,gz;
  int8_t gx_os,gy_os,gz_os;
  
 
  mpu.getRotation(&gx,&gy,&gz);
  gx_os=mpu.getXGyroOffset();
  gy_os=mpu.getYGyroOffset();
  gz_os=mpu.getZGyroOffset();
  
  #ifdef DEBUG_PRINTING
  Serial.println("Calibrating gyro, assuming it is laying still and flat");
  Serial.println("Gyro readings:");
  Serial.println(gx);
  Serial.println(gy);
  Serial.println(gz);
  Serial.println("Current gyro offsets");
  Serial.println(gx_os);
  Serial.println(gy_os);
  Serial.println(gz_os);
  
  Serial.println("Setting all offsets to 0");
  #endif
  mpu.setXGyroOffset(0);
  mpu.setYGyroOffset(0);
  mpu.setZGyroOffset(0);
  
  delay(100);
  mpu.getRotation(&gx,&gy,&gz);
  #ifdef DEBUG_PRINTING
  Serial.println("Gyro readings:");
  Serial.println(gx);
  Serial.println(gy);
  Serial.println(gz);
  
  Serial.println("Setting new offsets");
  #endif
  mpu.setXGyroOffset(gx/4);
  mpu.setYGyroOffset(gy/4);
  mpu.setZGyroOffset(gz/4);
  #ifdef DEBUG_PRINTING
  Serial.println("Current gyro offsets");
  Serial.println(mpu.getXGyroOffset());
  Serial.println(mpu.getYGyroOffset());
  Serial.println(mpu.getZGyroOffset());
  #endif
  mpu_calibration_complete=true;
}

//arm the ESC
void arm(){
  pitch.writeMicroseconds(ESC_MIN);
  roll.writeMicroseconds(ESC_MIN);
  yaw.writeMicroseconds(ESC_MIN);
  throttle.writeMicroseconds(ESC_MIN);
  delay(ESC_ARM_DELAY);
}

void initESCs(){
  pitch.attach(FC_PITCH);
  roll.attach(FC_ROLL);
  yaw.attach(FC_YAW);
  throttle.attach(FC_THROTT);
  arm();
}
